/*
  forward.c

  $Id: forward.c,v 1.2 1999/11/29 03:31:20 everton Exp $
 */

#include <sys/time.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <errno.h>
#include <string.h>
#include <unistd.h>
#include <arpa/inet.h>
#include <pthread.h>
#include <syslog.h>
#include "msignal.h"
#include "portfwd.h"
#include "forward.h"
#include "util.h"
#include "solve.h"
#include "msignal.h"
#include "host_map.hpp"
#include "iterator.hpp"

/*
 * Returns socket descriptor on success; -1 on failure.
 */
int tcp_match(const vector<host_map*> *map_list, const struct ip_addr *ip, int port)
{
  int rsd;
  iterator<vector<host_map*>,host_map*> it(*map_list);
  for (it.start(); it.cont(); it.next()) {
    rsd = it.get()->tcp_match(ip, port);
    if (rsd != -1)
      return rsd;
  }
  return -1;
}

void m_close(int fd) 
{
  if (shutdown(fd, 0)) 
    syslog(LOG_ERR, "shutdown() on socket failed: %m");
  if (close(fd))
    syslog(LOG_ERR, "close() on socket failed: %m");
}

int buf_copy(int src_fd, int trg_fd)
{
  char buf[BUF_SZ];
  int rd = read(src_fd, buf, BUF_SZ);
  if (!rd)
    return -1;
  if (rd < 0) {
    syslog(LOG_ERR, "copy: Failure reading from socket: %m");
    return -1;
  }
  int wr = write(trg_fd, buf, rd);
  if (!wr)
    return -1;
  if (wr < rd) {
    syslog(LOG_ERR, "copy: Partial write to socket: %m");
    return -1;
  }
  return 0;
}

void tcp_forward(proto_t proto, int port, vector<host_map*> *map_list)
{
  int sd = socket(AF_INET, SOCK_STREAM, get_protonumber(proto));
  if (sd == -1) {
    syslog(LOG_ERR, "Can't create TCP socket: %s\n", strerror(errno));
    return;
  }
    
  struct sockaddr_in sa;
  sa.sin_family      = AF_INET;
  sa.sin_port        = htons(port);
  sa.sin_addr.s_addr = htonl(INADDR_ANY);
  memset((char *) sa.sin_zero, 0, sizeof(sa.sin_zero));

  if (bind(sd, (struct sockaddr *) &sa, sizeof(sa))) {
    syslog(LOG_ERR, "Can't bind TCP socket: %s: %s:%d\n", strerror(errno), inet_ntoa(sa.sin_addr), port);
    close(sd);
    return;
  }

  if (listen(sd, 3)) {
    syslog(LOG_ERR, "Can't listen TCP socket: %s\n", strerror(errno));
    close(sd);
    return;
  }

  const int MAX_FD = 256;
  int dest_fd[MAX_FD];

  fd_set fds, tmp_fds;
  int fd_set_len = sizeof(fd_set);
  FD_ZERO(&fds);
  FD_SET(sd, &fds);
  int maxfd = sd + 1;
  int nd;

  ONVERBOSE(syslog(LOG_DEBUG, "Waiting TCP connection on %s:%d", inet_ntoa(sa.sin_addr), port));
  for (;;) { /* forever */
    /*
     * Restores tmp_fds from fds.
     */
    memcpy(&tmp_fds, &fds, fd_set_len);

    /*
     * Wait for event: connection on sd or data on anyone else.
     */
    nd = select(maxfd, &tmp_fds, 0, 0, 0);
    if (nd < 0) {
      if (nd == -1)
	syslog(LOG_ERR, "TCP select() failed: %m");
      break;
    }

    /*
     * Let's handle sd here (mother socket).
     */
    if (FD_ISSET(sd, &tmp_fds)) {

      --nd;

      /*
       * Clear sd so it's ignored later.
       */
      FD_CLR(sd, &tmp_fds); 

      struct sockaddr_in cli_sa;
      unsigned int cli_sa_len = sizeof(cli_sa);
      int csd = accept(sd, (struct sockaddr *) &cli_sa, &cli_sa_len);
      if (csd < 0)
	syslog(LOG_ERR, "Can't accept TCP socket: %m");
      else {
	int cli_port = ntohs(cli_sa.sin_port);
	ONVERBOSE(syslog(LOG_DEBUG, "TCP connection from %s:%d", inet_ntoa(cli_sa.sin_addr), cli_port));

	/*
	 * Connect to destination.
	 */
	struct ip_addr ip;
	ip.addr = (char *) &(cli_sa.sin_addr.s_addr);
	ip.len  = addr_len;
	  
	int rsd = tcp_match(map_list, &ip, cli_port);
	if (rsd == -1) {
	  ONVERBOSE(syslog(LOG_DEBUG, "Address miss"));
	  m_close(csd);
	}
	else {
	  ONVERBOSE(syslog(LOG_DEBUG, "Address match"));

	  if ((csd >= MAX_FD) || (rsd >= MAX_FD)) {
	    syslog(LOG_ERR, "Destination socket descriptors overflow");
	    m_close(csd);
	    m_close(rsd);
	  }
	  else {
	    /*
	     * Add pair of communicating sockets.
	     */
	    FD_SET(csd, &fds);
	    FD_SET(rsd, &fds);
	    maxfd = MAX(csd + 1, MAX(rsd + 1, maxfd)); 
	  
	    /*
	     * Save peers so they can be remembered later.
	     */
	    dest_fd[csd] = rsd;
	    dest_fd[rsd] = csd;
	  }
	}
      }

    } /* sd (mother socket) handled */

    /*
     * Now all other sockets.
     */
    for (int src_fd = 0; nd; ++src_fd)
      if (FD_ISSET(src_fd, &tmp_fds)) {
	--nd;
	int trg_fd = dest_fd[src_fd];
	/*
	 * Copy data.
	 */
	int fail = buf_copy(src_fd, trg_fd);
	if (fail) {
	  /*
	   * Remove pair of communicating sockets.
	   */
	  FD_CLR(src_fd, &fds);
	  FD_CLR(trg_fd, &fds);
	  m_close(src_fd);
	  m_close(trg_fd);
	}
      }

  } /* main loop */

}

void do_udp_forward(struct sockaddr_in *sa, vector<host_map*> *map_list, const char *buf, int buf_len)
{
  int                port      = ntohs(sa->sin_port);
  struct ip_addr     ip;

  ONVERBOSE(syslog(LOG_DEBUG, "TCP packet from: %s:%d\n", 
		   inet_ntoa(sa->sin_addr), 
		   ntohs(sa->sin_port)));

  ip.addr = (char *) &(sa->sin_addr.s_addr);
  ip.len  = addr_len;

  iterator<vector<host_map*>,host_map*> it(*map_list);
  for (it.start(); it.cont(); it.next())
    if (it.get()->udp_match(&ip, port, buf, buf_len))
      break;
}

void udp_forward(proto_t proto, int port, vector<host_map*> *map_list)
{
  int sd = socket(AF_INET, SOCK_DGRAM, get_protonumber(proto));
  if (sd == -1) {
    syslog(LOG_ERR, "Can't create UDP socket: %s\n", strerror(errno));
    return;
  }
    
  struct sockaddr_in sa;
  sa.sin_family      = AF_INET;
  sa.sin_port        = htons(port);
  sa.sin_addr.s_addr = htonl(INADDR_ANY);
  memset((char *) sa.sin_zero, 0, sizeof(sa.sin_zero));

  if (bind(sd, (struct sockaddr *) &sa, sizeof(sa))) {
    syslog(LOG_ERR, "Can't bind UDP socket: %s: %s:%d\n", strerror(errno), inet_ntoa(sa.sin_addr), port);
    close(sd);
    return;
  }

  char buf[BUF_SZ];
  int  rd;
  struct sockaddr_in cli_sa;
  unsigned int cli_sa_len = sizeof(cli_sa);

  for (;;) { /* forever */
    ONVERBOSE(syslog(LOG_DEBUG, "Waiting UDP packet on %s:%d", inet_ntoa(cli_sa.sin_addr), port));

    rd = recvfrom(sd, buf, BUF_SZ, 0, (struct sockaddr *) &cli_sa, &cli_sa_len);
    if (rd == -1) {
      syslog(LOG_ERR, "Can't receive UDP packet: %m");
      break;
    }

    do_udp_forward(&cli_sa, map_list, buf, rd);
  }

  if (close(sd))
    syslog(LOG_WARNING, "close on socket failed: %m");
}

