/*
  util.c

  $Id: util.c,v 1.1 1999/11/27 03:00:49 everton Exp $
 */

#include <stdarg.h>
#include <string.h>
#include <time.h>
#include <string.h>
#include <unistd.h>
#include <sys/ioctl.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <sys/time.h>
#include <sys/resource.h>
#include <sys/file.h>
#include <errno.h>
#include <stdio.h>
#include <syslog.h>

int verbose_mode = 0; /* See ONVERBOSE in util.h */

const char *prog_name = 0;

void set_prog_name(const char *name)
{
  prog_name = name;
}

const char *get_prog_name()
{
  return prog_name ? prog_name : "?";
}

char* safe_strdup(const char* str)
{
  const char* fmt  = "%s: %s\n";
  const char* func = "safe_strdup";

  if (!str) {
    syslog(LOG_EMERG, fmt, func, "null string pointer\n");
    exit(1);
  }
  char* s = strdup(str);
  if (!s) {
    syslog(LOG_EMERG, fmt, func, "could not allocate string\n");
    exit(1);
  }

  return s;
}

void safe_strcpy(char *trg, const char *src, int trg_buf_size)
{
  for (char *past_end = trg + trg_buf_size - 1; trg < past_end; ++src, ++trg)
    {
      char c = *src;
      *trg = c;
      if (!c)
        return;
    }
  *trg = '\0';
}

int daemonize()
{
  /*
   * Desconecta terminal de controle.
   */
  const char *tty_path = "/dev/tty";
  int tty = open(tty_path, O_NOCTTY);
  if (tty == -1) {
    syslog(LOG_ERR, "daemonize(): can't open tty: %s: %s\n", tty_path, strerror(errno));
    return -1;
  }
  if (ioctl(tty, TIOCNOTTY)) {
    syslog(LOG_ERR, "daemonize(): can't detach terminal: %s\n", strerror(errno));
    return -1;
  }
  if (close(tty)) {
    syslog(LOG_ERR, "daemonize(): can't close tty: %s: %s\n", tty_path, strerror(errno));
    return -1;
  }

  /*
   * Garante que futuros open()s nao vao alocar um terminal de controle.
   */
  pid_t pid = fork();
  if (pid) {
    /* Parent */
    if (pid < 0) {
      syslog(LOG_ERR, "daemonize(): fork() failed\n");
      return -1;
    }
    exit(0);
  }

  /*
   * Obtem numero maximo de arquivos que podemos abrir.
   */
  struct rlimit rl;
  if (getrlimit(RLIMIT_NOFILE, &rl) < 0) {
    syslog(LOG_ERR, "daemonize(): can't get file limit: %s\n", strerror(errno));
    return -1;
  }
  
  /*
   * Fecha todos os descritores abertos.
   */
  int fd;
  if (rl.rlim_max == RLIM_INFINITY)
    rl.rlim_max = 1024;
  for (fd = 0; fd < rl.rlim_max; ++fd)
    close(fd);
  errno = 0; /* close() em fd invalido */
  
  /*
   * Libera as permissoes dos arquivos criados pelo processo.
   */
  umask(0);

  /*
   * Muda para o diretorio raiz para nao atrapalhar umount.
   */
  if (chdir("/") < 0) {
    syslog(LOG_ERR, "daemonize(): can't chdir to / (root): %s\n", strerror(errno));
    return -1;
  }

  /*
   * Conecta stdin, stdout e stderr a /dev/null.
   */
  const char *dev_null = "/dev/null";
  fd = open(dev_null, O_RDWR);
  if (fd) {
    if (fd < 0)
      syslog(LOG_ERR, "daemonize(): error opening %s: %s\n", dev_null, strerror(errno));
    else
      syslog(LOG_ERR, "daemonize(): opening %s: expected fd 0, got %d\n", dev_null, fd);
    return -1;
  }

  fd = dup(0);
  if (fd != 1) {
    if (fd < 0)
      syslog(LOG_ERR, "daemonize(): dup(0) failed: %s", strerror(errno));
    else
      syslog(LOG_ERR, "daemonize(): dup(0) expected fd 1, got %d\n", fd);
    return -1;
  }

  fd = dup(0);
  if (fd != 2) {
    if (fd < 0)
      syslog(LOG_ERR, "daemonize(): dup(0) failed: %s", strerror(errno));
    else
      syslog(LOG_ERR, "daemonize(): dup(0) expected fd 2, got %d\n", fd);
    return -1;
  }

  return 0;
}


/* eof: util.c */
